"use strict";
/*
    Senselogs - Simple, fast, dynamic logging for serverless
 */
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _SenseLogs_defaultFilter, _SenseLogs_destinations, _SenseLogs_filter, _SenseLogs_options, _SenseLogs_override, _SenseLogs_name, _SenseLogs_redact, _SenseLogs_sample, _SenseLogs_timestamp, _SenseLogs_top;
Object.defineProperty(exports, "__esModule", { value: true });
const DefaultFilter = ['fatal', 'error', 'metrics', 'info', 'warn'];
class SenseLogs {
    /*
        Options:
            destination
            filter
            name
            redact
            timestamp
    */
    constructor(options = { destination: 'json' }, context = {}) {
        _SenseLogs_defaultFilter.set(this, void 0);
        _SenseLogs_destinations.set(this, void 0);
        _SenseLogs_filter.set(this, void 0);
        _SenseLogs_options.set(this, void 0);
        _SenseLogs_override.set(this, void 0);
        _SenseLogs_name.set(this, void 0);
        _SenseLogs_redact.set(this, void 0);
        _SenseLogs_sample.set(this, void 0);
        _SenseLogs_timestamp.set(this, void 0);
        _SenseLogs_top.set(this, void 0);
        if (!options.child) {
            __classPrivateFieldSet(this, _SenseLogs_options, Object.assign({}, options), "f");
            __classPrivateFieldSet(this, _SenseLogs_name, options.name || 'app', "f");
            __classPrivateFieldSet(this, _SenseLogs_redact, options.redact, "f");
            __classPrivateFieldSet(this, _SenseLogs_timestamp, options.timestamp || false, "f");
            __classPrivateFieldSet(this, _SenseLogs_top, this, "f");
            __classPrivateFieldSet(this, _SenseLogs_filter, {}, "f");
            __classPrivateFieldSet(this, _SenseLogs_defaultFilter, null, "f");
            __classPrivateFieldSet(this, _SenseLogs_override, {}, "f");
            __classPrivateFieldSet(this, _SenseLogs_sample, {}, "f");
            __classPrivateFieldSet(this, _SenseLogs_destinations, [], "f");
            let filter = options.filter || DefaultFilter;
            /* istanbul ignore next */
            if (typeof process != 'undefined' && process.env) {
                if (process.env.LOG_FILTER != null) {
                    filter = process.env.LOG_FILTER;
                }
                if (process.env.LOG_SAMPLE != null) {
                    let [percentage, filter] = process.env.LOG_SAMPLE.split(':');
                    this.setSample(filter, parseInt(percentage));
                }
                if (process.env.LOG_OVERRIDE != null) {
                    let [expire, filter] = process.env.LOG_OVERRIDE.split(':');
                    this.setOverride(filter, expire);
                }
            }
            this.setFilter(filter);
            __classPrivateFieldSet(this, _SenseLogs_defaultFilter, Object.keys(__classPrivateFieldGet(this, _SenseLogs_filter, "f")), "f");
            if (options.destination == 'console') {
                this.addDestination(new ConsoleDest());
            }
            else if (options.destination == 'capture') {
                this.addDestination(new CaptureDest());
            }
            else if (options.destination && typeof options.destination.write == 'function') {
                this.addDestination(options.destination);
            }
            else {
                //  Default to json
                this.addDestination(new JsonDest());
            }
            this.addUncaughtExceptions();
        }
        this.context = context;
    }
    /*
        Add a filter to enable specified channels.
    */
    addFilter(filter) {
        if (!filter) {
            return this;
        }
        if (!Array.isArray(filter)) {
            if (filter == 'default') {
                filter = __classPrivateFieldGet(this, _SenseLogs_defaultFilter, "f") || DefaultFilter;
            }
            else {
                filter = filter.split(',').map(f => f.trim());
            }
        }
        for (let chan of filter) {
            if (chan) {
                __classPrivateFieldGet(__classPrivateFieldGet(this, _SenseLogs_top, "f"), _SenseLogs_filter, "f")[chan] = true;
            }
        }
        return this;
    }
    setFilter(filter) {
        __classPrivateFieldSet(__classPrivateFieldGet(this, _SenseLogs_top, "f"), _SenseLogs_filter, {}, "f");
        if (filter) {
            return this.addFilter(filter);
        }
        return this;
    }
    getFilter() {
        return Object.keys(__classPrivateFieldGet(this, _SenseLogs_filter, "f"));
    }
    getOverride() {
        return __classPrivateFieldGet(this, _SenseLogs_override, "f");
    }
    getSample() {
        return __classPrivateFieldGet(this, _SenseLogs_sample, "f");
    }
    /*
        Define a limited duration override filter. If filter is null, clear overrides.
    */
    setOverride(filter, expire) {
        if (!filter) {
            __classPrivateFieldSet(this, _SenseLogs_override, {}, "f");
            return this;
        }
        if (!expire) {
            expire = Date.now();
        }
        if (!Array.isArray(filter)) {
            filter = filter.split(',').map(f => f.trim());
        }
        for (let chan of filter) {
            if (chan) {
                __classPrivateFieldGet(__classPrivateFieldGet(this, _SenseLogs_top, "f"), _SenseLogs_override, "f")[chan] = expire;
            }
        }
        return this;
    }
    /*
        Define a sampling filter to apply to a percentage of requests. If sample is null, clear samples.
    */
    setSample(filter, rate) {
        if (!filter) {
            __classPrivateFieldSet(this, _SenseLogs_sample, {}, "f");
            return this;
        }
        if (rate == null) {
            rate = 0;
        }
        if (rate > 100) {
            rate = 100;
        }
        if (!Array.isArray(filter)) {
            filter = filter.split(',').map(f => f.trim());
        }
        if (rate > 0) {
            for (let chan of filter) {
                if (chan) {
                    __classPrivateFieldGet(__classPrivateFieldGet(this, _SenseLogs_top, "f"), _SenseLogs_sample, "f")[chan] = { count: 0, total: 100 / rate };
                }
            }
        }
        else {
            for (let chan of filter) {
                if (chan) {
                    __classPrivateFieldGet(__classPrivateFieldGet(this, _SenseLogs_top, "f"), _SenseLogs_sample, "f")[chan] = null;
                }
            }
        }
        return this;
    }
    /*
        Create a child logger. This inherits the context and channels.
    */
    child(context) {
        context = context ? Object.assign({}, this.context, context) : this.context;
        let log = new SenseLogs({ child: true }, context);
        __classPrivateFieldSet(log, _SenseLogs_top, __classPrivateFieldGet(this, _SenseLogs_top, "f"), "f");
        return log;
    }
    /*
        Add a new logger destination. Multiple destinations are supported.
    */
    addDestination(dest) {
        __classPrivateFieldGet(__classPrivateFieldGet(this, _SenseLogs_top, "f"), _SenseLogs_destinations, "f").push(dest);
        return this;
    }
    setDestination(dest) {
        __classPrivateFieldSet(__classPrivateFieldGet(this, _SenseLogs_top, "f"), _SenseLogs_destinations, [], "f");
        __classPrivateFieldGet(__classPrivateFieldGet(this, _SenseLogs_top, "f"), _SenseLogs_destinations, "f").push(dest);
        return this;
    }
    /*
        Add (blend) a contexts into the context for this logger
    */
    addContext(contexts) {
        if (!Array.isArray(contexts)) {
            contexts = [contexts];
        }
        for (let context of contexts) {
            this.context = Object.assign(this.context, context);
        }
        return this;
    }
    /*
        Clear this loggers context
    */
    clearContext() {
        this.context = {};
        return this;
    }
    /*
        Determine if a channel should emit a log message
    */
    shouldLog(chan) {
        let top = __classPrivateFieldGet(this, _SenseLogs_top, "f");
        if (__classPrivateFieldGet(top, _SenseLogs_filter, "f")[chan] == null) {
            if (__classPrivateFieldGet(top, _SenseLogs_override, "f")[chan] && __classPrivateFieldGet(top, _SenseLogs_override, "f")[chan] < Date.now()) {
                __classPrivateFieldGet(top, _SenseLogs_override, "f")[chan] = null;
            }
            if (__classPrivateFieldGet(top, _SenseLogs_override, "f")[chan] == null) {
                let sample = __classPrivateFieldGet(top, _SenseLogs_sample, "f")[chan];
                if (sample == null) {
                    return false;
                }
                if (++sample.count < sample.total) {
                    return false;
                }
                sample.count = 0;
            }
        }
        return true;
    }
    /*
        Process a log message and optionally emit
    */
    process(chan, message, context) {
        if (!this.shouldLog(chan)) {
            return;
        }
        let exception;
        if (context instanceof Error) {
            exception = context;
            context = {};
        }
        else {
            context = Object.assign({}, context);
        }
        context['@chan'] = chan;
        context['@module'] = context['@module'] || __classPrivateFieldGet(__classPrivateFieldGet(this, _SenseLogs_top, "f"), _SenseLogs_name, "f");
        if (__classPrivateFieldGet(this, _SenseLogs_timestamp, "f")) {
            context.timestamp = new Date();
        }
        if (context['@stack'] === true) {
            try {
                context['@stack'] = (new Error('stack')).stack.split('\n')[3].trim().replace(/^.*webpack:\/|:[0-9]*\)$/g, '');
            }
            catch (err) { }
        }
        if (context.message) {
            context['@message'] = context.message;
        }
        if (message instanceof Error) {
            exception = message;
            message = exception.message;
        }
        else if (context.err instanceof Error) {
            exception = context.err;
            delete context.err;
        }
        else if (typeof message != 'string') {
            message = JSON.stringify(message);
        }
        if (exception) {
            context['@exception'] = exception;
            message = message || exception.message;
        }
        context.message = message;
        context = Object.assign({}, this.context, context);
        if (__classPrivateFieldGet(this, _SenseLogs_redact, "f")) {
            context = __classPrivateFieldGet(this, _SenseLogs_redact, "f").call(this, context);
            if (!context)
                return;
        }
        this.write(context);
    }
    data(message, context) { this.process('data', message, context); }
    debug(message, context) { this.process('debug', message, context); }
    error(message, context) { this.process('error', message, context); }
    fatal(message, context) { this.process('fatal', message, context); }
    info(message, context) { this.process('info', message, context); }
    trace(message, context) { this.process('trace', message, context); }
    warn(message, context) { this.process('warn', message, context); }
    /*
        Write a log message to the configured destinations
    */
    write(context) {
        for (let dest of __classPrivateFieldGet(__classPrivateFieldGet(this, _SenseLogs_top, "f"), _SenseLogs_destinations, "f")) {
            dest.write(__classPrivateFieldGet(this, _SenseLogs_top, "f"), context);
        }
    }
    flush(context) {
        let buffer;
        for (let dest of __classPrivateFieldGet(__classPrivateFieldGet(this, _SenseLogs_top, "f"), _SenseLogs_destinations, "f")) {
            buffer = dest.flush(__classPrivateFieldGet(this, _SenseLogs_top, "f"), context);
        }
        return buffer;
    }
    /*
        Convenience API to emit a log message for a given channel.
    */
    emit(chan, message, context) {
        this.process(chan, message, context);
    }
    /*
        Emit a CloudWatch metic using the CloudWatch EMF format.

        metrics('MyCompany/MyApp', {UserSessions: 1}, dimensions)
    */
    metrics(namespace, values, dimensions = [[]]) {
        if (!__classPrivateFieldGet(__classPrivateFieldGet(this, _SenseLogs_top, "f"), _SenseLogs_filter, "f").metrics) {
            return;
        }
        /* istanbul ignore next */
        if (!namespace || !values) {
            throw new Error('Missing namespace or values');
        }
        let keys = Object.keys(values).filter(v => dimensions[0].indexOf(v) < 0);
        let metrics = keys.map(v => { return { Name: v }; });
        let context = {
            '@chan': 'metrics',
            '@namespace': namespace,
            '@metrics': keys,
            message: `Metrics for ${namespace} ` + JSON.stringify(Object.assign({
                _aws: {
                    Timestamp: Date.now(),
                    CloudWatchMetrics: [{
                            Dimensions: dimensions,
                            Namespace: namespace,
                            Metrics: metrics,
                        }]
                },
            }, values))
        };
        for (let dest of __classPrivateFieldGet(__classPrivateFieldGet(this, _SenseLogs_top, "f"), _SenseLogs_destinations, "f")) {
            dest.write(__classPrivateFieldGet(this, _SenseLogs_top, "f"), context);
        }
    }
    addUncaughtExceptions() {
        let self = this;
        if (typeof window != 'undefined') {
            /* istanbul ignore next */
            global.onerror = function (message, module, line, column, err) {
                self.error(message, err);
            };
            /* istanbul ignore next */
            global.onunhandledrejection = (rejection) => {
                let message = `Unhandled promise rejection : ${rejection.message}`;
                if (rejection && rejection.reason && rejection.reason.stack) {
                    message += `\r${rejection.reason.stack}`;
                }
                self.error(message);
            };
        }
    }
    addNodeExceptions() {
        let self = this;
        /* istanbul ignore next */
        if (typeof process != 'undefined') {
            process.on("uncaughtException", function (err) {
                self.error('Uncaught exception', err);
            });
        }
    }
}
exports.default = SenseLogs;
_SenseLogs_defaultFilter = new WeakMap(), _SenseLogs_destinations = new WeakMap(), _SenseLogs_filter = new WeakMap(), _SenseLogs_options = new WeakMap(), _SenseLogs_override = new WeakMap(), _SenseLogs_name = new WeakMap(), _SenseLogs_redact = new WeakMap(), _SenseLogs_sample = new WeakMap(), _SenseLogs_timestamp = new WeakMap(), _SenseLogs_top = new WeakMap();
/*
    Log output in pure JSON
*/
class JsonDest {
    write(log, context) {
        let chan = context['@chan'];
        if (chan == 'metrics') {
            process.stdout.write(context.message + '\n');
        }
        else {
            process.stdout.write(JSON.stringify(context) + '\n');
        }
    }
}
/*
    Capture Destination
*/
class CaptureDest {
    constructor() {
        this.buffer = [];
    }
    flush() {
        let buffer = this.buffer;
        this.buffer = [];
        return buffer;
    }
    write(log, context) {
        let chan = context['@chan'];
        if (chan == 'metrics') {
            this.buffer.push(context.message);
        }
        else {
            this.buffer.push(context);
        }
    }
}
/*
    Simple destination to the console. Uses abbreviated time format.
*/
class ConsoleDest {
    write(log, context) {
        let message = context.message;
        let module = context['@module'];
        let time = this.getTime();
        let chan = context['@chan'];
        let exception = context['@exception'];
        if (exception) {
            process.stderr.write(`${time}: ${module}: ${chan}: ${message}: ${exception.message}` + '\n');
            process.stderr.write(exception.stack + '\n');
            process.stderr.write(JSON.stringify(context, null, 4) + '\n');
        }
        else if (chan == 'error') {
            process.stderr.write(`${time}: ${module}: ${chan}: ${message}` + '\n');
            if (Object.keys(context).length > 3) {
                process.stderr.write(JSON.stringify(context, null, 4) + '\n');
            }
        }
        else if (chan == 'metrics') {
            process.stdout.write(context.message + '\n');
        }
        else if (chan == 'trace') {
            process.stdout.write(`${time}: ${module}: ${chan}: ${message}` + '\n');
            process.stdout.write(JSON.stringify(context, null, 4) + '\n');
        }
        else {
            process.stdout.write(`${time}: ${module}: ${chan}: ${message}` + '\n');
            if (Object.keys(context).length > 3) {
                //  More than: message, @module and @chan
                process.stdout.write(JSON.stringify(context, null, 4) + '\n');
            }
        }
    }
    //  Abbreviated timea for console destination
    getTime() {
        let now = new Date();
        return `${this.zpad(now.getHours(), 2)}:${this.zpad(now.getMinutes(), 2)}:${this.zpad(now.getSeconds(), 2)}`;
    }
    zpad(n, size) {
        let s = n + '';
        while (s.length < size)
            s = '0' + s;
        return s;
    }
}
